
import http.client
from urllib.parse import urlparse, urlencode, quote_plus


class HTTPClient(object):

    def __init__(self, base_url):
        self.parsed_url = urlparse(base_url)
        if self.parsed_url.scheme == 'https':
            self.conn = http.client.HTTPSConnection(self.parsed_url.netloc)
        else:
            self.conn = http.client.HTTPConnection(self.parsed_url.netloc)

    def get(self, url):
        self.conn.request('GET', url)
        response = conn.getresponse()
        return response.read()

    def post(self, url, data):
        encoded_data = urlencode(data, quote_via=quote_plus)
        headers = {'Content-type': 'application/x-www-form-urlencoded',
                   'Accept': 'text/plain'}
        self.conn.request('POST', url, encoded_data, headers)
        response = conn.getresponse()
        return response.read()




import os
from configparser import SafeConfigParser


class Config(object):
    def __init__(self, path=None):
        self.path = path
        self.default_paths = [
            os.path.expanduser('~/'),
            os.path.join(os.path.dirname(__file__),'../etc')
        ]
        self.config = {}

    @property
    def filename(self):
        return 'stamper.conf'

    @property
    def sections(self):
        return['stamps', 'sum', 'charts', 'collector']

    def look_for_config(self):
        """
        Look for the configuration file, following the path names listed in
        the default paths
        """
        for path in self.default_paths:
            full_path = os.path.join(path, self.filename)
            if os.path.isfile(full_path):
                self.path = full_path
                return full_path
        # if we reach here, self.path will be still None, no valid
        # config files were found, and so we raise an exception
        raise IOError('ERROR - Can not find ' + self.filename + \
                      ' in your environment')

    def load(self):
        if not self.path:
            full_path = self.look_for_config()
        parser = SafeConfigParser()
        parser.read(full_path)
        for section in self.sections:
            self.config[section] = {}
            for name, value in parser.items(section):
                self.config[section][name] = value
        return self.config

    def get(self, section, parameter=None):
        if section not in self.config.keys():
            # perhaps the config hasn't been loaded yet
            self.load()

        if section not in self.sections:
            raise IndexError('ERROR - ' + section + \
                             ' is not one of the available sections: ' + \
                             ', '.join(self.sections))

        if parameter:
            return self.config[section][parameter]

        return self.config[section]




import json
import os.path
from datetime import datetime, timedelta
from os import symlink, remove, makedirs
from collections import OrderedDict
from operator import itemgetter
import pygal

from .http import HTTPClient
from .config import Config
from .filters import DateFilter


class Stamper(object):

    def __init__(self, config_file=None):
        self.config = Config(config_file)
        self.stamps_file = os.path.expanduser(
            self.config.get('stamps', 'path'))
        self.charts_dir = os.path.expanduser(
            self.config.get('charts', 'path'))
        self.date_format = self.config.get('stamps', 'date_format')
        self.time_format = self.config.get('stamps', 'time_format')
        self.datetime_format = self.config.get('stamps', 'datetime_format')
        self.wants_seconds = self.config.get('stamps', 'wants_seconds')
        self.hours_day = int(self.config.get('sum', 'hours_day'))
        self.collector = self.config.get('collector')
        self.ensure_stamps_file()
        self.ensure_charts_dir()
        self.stamps = []
        self.date_filter = DateFilter(self.date_format)

    def __json_load(self, filename):
        """
        Load the stamps from a file in json format, returning
        the parsed list.
        """
        with open(filename, 'r') as stamps_file:
            try:
                stamps = json.load(stamps_file)
            except ValueError:
                stamps = []
        return stamps

    def remove_duplicates(self):
        """
        Remove duplicated stamps from the stamps list
        """
        stamps = [dict(t) for t in set(
            [tuple(d.items()) for d in self.stamps])]
        self.stamps = stamps

    def ensure_stamps_file(self):
        if not os.path.exists(self.stamps_file):
            with open(self.stamps_file, 'w') as stamps_file:
                stamps_file.write('')

    def ensure_charts_dir(self):
        if not os.path.exists(self.charts_dir):
            makedirs(self.charts_dir)

    def load_stamps(self):
        self.stamps = self.__json_load(self.stamps_file)

    def sort_stamps(self):
        """
        Sort all the stamps by start and end dates
        """
        self.stamps = sorted(self.stamps, key=itemgetter('start', 'end'))

    def save_stamps(self):
        with open(self.stamps_file, 'w') as stamps_file:
            json.dump(self.stamps, stamps_file, indent=4)

    def stamp(self, start, end, customer, action):
        self.stamps.append({
            'start': start,
            'end': end,
            'customer': customer,
            'action': action,
        })

    def last_stamp(self, n=1):
        """
        return the stamp in position -n, that is, starting from the latest one
        and going back N positions in the list of stamps
        """
        if not self.stamps:
            return None
        return self.stamps[-n]

    def worktime(self, start, end):
        worktime = (datetime.strptime(end, self.datetime_format) -
                    datetime.strptime(start, self.datetime_format))
        return worktime.seconds

    @property
    def customers(self):
        customers = []
        for stamp in self.stamps:
            if stamp['customer'] not in customers:
                customers.append(stamp['customer'])
        customers.remove(None)
        return customers

    def totals(self, filter_from=None, filter_to=None, filter_descr=None):
        totals = {}
        for stamp in self.stamps:
            customer = stamp['customer']
            # customer will be None for "start" stamps, having no end time
            if customer:
                start = datetime.strptime(stamp['start'], self.datetime_format)
                end = datetime.strptime(stamp['end'], self.datetime_format)
                if filter_from and start < filter_from:
                    # if there is a filter setting a starting date for the
                    # report and the current stamp is from an earlier date, do
                    # not add it to the totals
                    continue
                if filter_to and start > filter_to:
                    # similar for the end date
                    continue
                if filter_descr and filter_descr not in stamp['action']:
                    continue
                if customer not in totals:
                    totals[customer] = 0
                totals[customer] += self.worktime(stamp['start'], stamp['end'])
        return totals

    def details(self, filter_customer=None, filter_from=None, filter_to=None,
                filter_descr=None):
        details = OrderedDict()
        totals = OrderedDict()
        total_customer = OrderedDict()
        for stamp in self.stamps:
            customer = stamp['customer']
            if customer:
                if filter_customer and customer != filter_customer:
                    # we are getting the details for only one customer, if this
                    # stamp is not for that customer, simply move on and ignore
                    # it
                    continue
                start = datetime.strptime(stamp['start'], self.datetime_format)
                start_day = start.strftime('%Y-%m-%d')
                end = datetime.strptime(stamp['end'], self.datetime_format)
                if filter_from and start < filter_from:
                    # if there is a filter setting a starting date for the
                    # report and the current stamp is from an earlier date, do
                    # not add it to the totals
                    continue
                if filter_to and start > filter_to:
                    # similar for the end date
                    continue
                if filter_descr and filter_descr not in stamp['action']:
                    continue
                # avoid "start" stamps
                if start_day not in details:
                    details[start_day] = []
                if start_day not in totals:
                    totals[start_day] = 0
                worktime = self.worktime(stamp['start'], stamp['end'])
                str_worktime = str(timedelta(seconds=worktime))
                if not self.wants_seconds or self.wants_seconds == 'false':
                    # remove the seconds part from the string representation
                    str_worktime = ':'.join(str_worktime.split(':')[:-1])
                details[start_day].append(
                    '%(worktime)s %(customer)s %(action)s' % {
                        'worktime': str_worktime,
                        'customer': customer,
                        'action': stamp['action']
                    })
                totals[start_day] += worktime
                if start_day not in total_customer:
                    total_customer[start_day] = {}
                if customer not in total_customer[start_day]:
                    total_customer[start_day][customer] = 0
                total_customer[start_day][customer] += worktime
        for day in totals:
            totals[day] = str(timedelta(seconds=totals[day]))
            if not self.wants_seconds or self.wants_seconds == 'false':
                # remove the seconds part from the string representation
                totals[day] = ':'.join(totals[day].split(':')[:-1])
        return details, totals, total_customer

    def timeline(self, customer=None, stamp_filter=None, filter_descr=None):
        filter_from, filter_to = self.date_filter.validate(stamp_filter)
        for stamp in self.stamps:
            start = datetime.strptime(stamp['start'], self.datetime_format)
            start_day = start.strftime('%Y-%m-%d')
            if filter_from and start < filter_from:
                # if there is a filter setting a starting date for the
                # report and the current stamp is from an earlier date, do
                # not add it to the totals
                continue
            if filter_to and start > filter_to:
                # similar for the end date
                continue
            if filter_descr and stamp['action']:
                if filter_descr not in stamp['action']:
                    continue
            if not stamp['customer']:
                if customer is None:
                    print(stamp['start'] + ' start')
            else:
                if customer and customer != stamp['customer']:
                    continue
                if customer:
                    print(stamp['start'] + ' start')
                print(' '.join([stamp['end'],
                                stamp['customer'],
                                stamp['action']]))

    def graph_stamps(self, customer=None, stamp_filter=None, filter_descr=None):
        """
        Generate charts with information from the stamps
        """
        filter_from, filter_to = self.date_filter.validate(stamp_filter)
        chart = pygal.Bar(title='Work hours per day',
                          range=(0, self.hours_day),
                          x_title='Days',
                          y_title='Work hours',
                          x_label_rotation=45)

        details, totals, totals_customers = self.details(customer,
                                                         filter_from,
                                                         filter_to,
                                                         filter_descr)
        days = []
        values = {}
        for c in self.customers:
            values[c] = []

        found = []

        for day in details:
            for c in values:
                seconds = totals_customers[day].get(c, 0)
                if seconds and c not in found:
                    found.append(c)
                human = timedelta(seconds=seconds).__str__()
                values[c].append({'value': seconds/60.00/60.00,
                                  'label': day + ': ' + human})
            days.append(day)
        chart.x_labels = map(str, days)

        if customer:
            chart.add(customer, values[customer])
        else:
            for c in found:
                chart.add(c, values[c])

        chart_name = 'chart-%s.svg' % datetime.today().strftime(
            '%Y-%m-%d_%H%M%S')
        chart_symlink = 'chart-latest.svg'
        chart_path = os.path.join(self.charts_dir, chart_name)
        chart_symlink_path = os.path.join(self.charts_dir, chart_symlink)

        chart.render_to_file(chart_path)
        print('Rendered chart: ' + chart_path)
        if os.path.islink(chart_symlink_path):
            remove(chart_symlink_path)
        symlink(chart_name, chart_symlink_path)
        print('Updated latest chart: ' + chart_symlink_path)

    def show_stamps(self, customer=None, stamp_filter=None, verbose=False,
        sum=False, filter_descr=None):
        filter_from, filter_to = self.date_filter.validate(stamp_filter)

        # If the user asks for verbose information, show it before the
        # totals (mimicing what the original stamp tool does)
        if verbose:
            details, totals, total_customer = self.details(customer,
                                                           filter_from,
                                                           filter_to,
                                                           filter_descr)
            for day in details:
                print('------ %(day)s ------' % {'day': day})
                for line in details[day]:
                    print(line)
                customer_day_totals = []
                for tc in total_customer[day]:
                    tc_total = str(timedelta(seconds=total_customer[day][tc]))
                    if not self.wants_seconds or self.wants_seconds == 'false':
                        # remove the seconds part from the string representation
                        tc_total = ':'.join(tc_total.split(':')[:-1])
                        # ensure hours show double digits
                        tc_total = tc_total.zfill(5)
                    customer_day_totals.append(tc+': '+tc_total)
                print(', '.join(customer_day_totals))
                if len(customer_day_totals) > 1:
                    # if there are multiple customers in the report, show the
                    # daily totals
                    print('daily total: %(total)s' % {'total': totals[day]})
            print('-'*79)

        # now calculate the totals and show them
        totals = self.totals(filter_from, filter_to, filter_descr)
        if customer:
            seconds=totals.get(customer, 0)
            total = timedelta(seconds=totals.get(customer, 0))
            if not self.wants_seconds or self.wants_seconds == 'false':
                # remove the seconds part from the string representation
                total = ':'.join(str(total).split(':')[:-1])
            print(' %(customer)s: %(total)s' % {'customer': customer,
                                                'total': total})
        else:
            for c in totals:
                seconds=totals[c]
                total = timedelta(seconds=totals[c])
                if not self.wants_seconds or self.wants_seconds == 'false':
                    # remove the seconds part from the string representation
                    total = ':'.join(str(total).split(':')[:-1])
                print(' %(customer)s: %(total)s' % {'customer': c,
                                                    'total': total})

        if sum:
            sum_tot = ''
            if totals:
                print('------ Totals ------' % {'day': day})
                for day, tot in totals.items():
                    print(' %(day)s: %(total)s' % {'day': day, 'total': tot})
                    sum_tot = "%(total)s %(new)s" % {
                        'total': sum_tot,
                        'new': total
                    }
                totalSecs, sec = divmod(seconds, 60)
                hr, min = divmod(totalSecs, 60)
                totalDays, remaining = divmod(seconds,
                                              (self.hours_day * 60 * 60))
                remainingMin, remainingSec = divmod(remaining, (60))
                remainingHr, remainingMin = divmod(remainingMin, (60))
                print('----- %d:%02d:%02d -----' % (hr, min, sec))
                print('--- %d days, remaining: %d:%02d (%d hours/day) ---' % (
                    totalDays, remainingHr, remainingMin, self.hours_day
                ))

    def remove_stamps(self, n=1):
        """
        Remove up to n stamps back, asking for confirmation before delete
        """
        for i in range(n):
            stamp = self.last_stamp()
            if not stamp['customer']:
                print(stamp['start'] + ' start')
            else:
                print(' '.join([stamp['end'],
                                stamp['customer'],
                                stamp['action']]))
            confirm = ''
            while confirm.lower() not in ['y', 'n']:
                confirm = raw_input('delete stamp? (y/n) ')
                confirm = confirm.lower()
            if confirm == 'y':
                self.stamps.pop()
            else:
                # if the user says no to the removal of an stamp, we cannot
                # keep deleting stamps after that one, as that could leave the
                # stamps in an inconsistent state.
                print('Aborting removal of stamps')
                break
        self.save_stamps()

    def import_stamps(self, filename):
        """
        Import the stamps from the given file into the main stamps list,
        merging them into the list (removing duplicated entries)
        """
        if not os.path.exists(filename):
            print('[error] ' + filename + 'does not exist')
            return
        if os.path.isdir(filename):
            print('[error] ' + filename + 'is a directory')
            return
        stamps = self.__json_load(filename)
        if not stamps:
            print('[warning] no stamps can be imported from ' + filename)
            return
        self.stamps.extend(stamps)
        self.remove_duplicates()
        self.sort_stamps()
        self.save_stamps()
        print('[warning] ' + str(len(stamps)) + ' stamps merged')
        print('[warning] remember to review the resulting stamps file')

    def push_stamps(self, customer=None, stamp_filter=None, filter_descr=None):
        filter_from, filter_to = self.date_filter.validate(stamp_filter)

        stamps = []
        for stamp in self.stamps:
            if stamp['customer']:
                if customer and customer != stamp['customer']:
                    continue
                start = datetime.strptime(stamp['start'], self.datetime_format)
                start_day = start.strftime('%Y-%m-%d')
                end = datetime.strptime(stamp['end'], self.datetime_format)
                if filter_from and start < filter_from:
                    # if there is a filter setting a starting date for the
                    # report and the current stamp is from an earlier date, do
                    # not add it to the totals
                    continue
                if filter_to and start > filter_to:
                    # similar for the end date
                    continue
                if filter_descr and filter_descr not in stamp['action']:
                    continue
                stamps.append(stamp)

        stamps = json.dumps(stamps, indent=4)
        http_client = HTTPClient(self.collector['base_url'])
        response = http_client.post(self.collector['login_path'],
                                    {'username': self.collector['user'],
                                    'password': self.collector['password']})
        # response is a json-encoded list of end-points from the collector api
        try:
            api = json.loads(response)
        except ValueError:
            print('[error] No valid api end points can be retrieved')
            return

        response = http_client.post(api['push'], {'stamps': stamps})

        # response is a json-encoded dict, containing lists of added, updated
        # and deleted stamps (on the other side)
        try:
            results = json.loads(response)
        except ValueError:
            print('[error] stamps pushed, can not retrieve results information')
            return

        # display information
        for k in results.keys():
            print('%(stamps)d stamps %(action)s' % {'stamps': len(results[k]),
                                                    'action': k})




import re
from datetime import datetime, date, timedelta


class DateFilter(object):

    """
    Filtering tools for the dates input provided by the user
    """

    def __init__(self, date_format):
        self.date_format = date_format

    def days(self, current_date, days, action='-'):
        """
        Add/substract the given number of days from the given current_date
        """
        delta = timedelta(days=days)
        if action == '-':
            new_date = current_date - delta
        else:
            new_date = current_date + delta
        new_date = new_date.replace(hour=0, minute=0, second=0)
        return new_date

    def days_ago(self, current_date, days):
        """
        Return the date N days ago from the current date
        """
        return self.days(current_date, days, '-')

    def days_future(self, current_date, days):
        """
        Return the date N days in the future from the current date
        """
        return self.days(current_date, days, '+')

    def months(self, current_date, months, action='-'):
        """
        Add/substract the given number of months from the given current_date
        """
        if action == '-':
            # start travelling in time, back to N months ago
            for n in range(months):
                current_date = current_date.replace(day=1) - timedelta(days=1)
        else:
            # start travelling in time, forward to N months
            for n in range(months):
                current_date = current_date.replace(day=1) + timedelta(days=1)

        # Now use the new year/month values + the current day to set the proper
        # date
        new_date = datetime(
            current_date.year, current_date.month, date.today().day)
        return new_date

    def months_ago(self, current_date, months):
        """
        Return the date N days ago from the current date
        """
        return self.months(current_date, months, '-')

    def months_future(self, current_date, months):
        """
        Return the date N days in the future from the current date
        """
        return self.months(current_date, months, '+')

    def parse_number_filter(self, number_filter, start_date=None,
                            future=False):
        """
        Given a numeric filter with the N[d|w|m|y] pattern, return both
        the number of days/months/years that apply as a filter, + the
        date N[d|w|m|y] ago from start_date (which defaults to today if
        None is passed).

        If future is True, it will return a date in the future, not from
        the past.
        """
        number = None
        filtered_date = None
        start_date = start_date or datetime.today()

        # set which methods use to process the date, going backwards into the
        # past as a default, unless we are told otherwise
        filter_days = self.days_ago
        filter_months = self.months_ago
        if future:
            filter_days = self.days_future
            filter_months = self.months_future

        if re.search(r'(\d+[dD]{1})', number_filter):
            number = int(number_filter.lower().replace('d', ''))
            filtered_date = filter_days(start_date, number)

        elif re.search(r'(\d+[wW]{1})', number_filter):
            number = int(number_filter.lower().replace('w', '')) * 7
            filtered_date = filter_days(start_date, number)

        elif re.search(r'(\d+[mM]{1})', number_filter):
            number = int(number_filter.lower().replace('m', ''))
            filtered_date = filter_months(start_date, number)

        elif re.search(r'(\d+[yY]{1})', number_filter):
            number = int(number_filter.lower().replace('y', ''))
            today = date.today()
            # by default assume going backwards into the past...
            year = today.year - number
            if future:
                #...unless told otherwise
                year = today.year + number
            filtered_date = datetime(year, today.month, today.day)

        return number, filtered_date

    def validate(self, stamp_filter):
        """
        Validate a given filter. Filters can have the following notation:

        - %Y-%m-%d: Times recorded at a given date

        - %Y-%m-%d--%Y-%m-%d: Times recorded between two dates

        - *%Y-%m-%d: Times recorded up to a  given date

        - %Y-%m-%d*: Times recorded from a given date

        - %Y-%m-%d+sN[d|w|m|y]: Times recorded since the given date up to
          N more days/weeks/months/years

        - N...N[d|w|m|y]: Times recorded N...N days/weeks/months/years ago

        Important: all date comparisons are made on datetime objects, using
        00:00 as the time (first second of the given day). This means that
        for range filters, the first day is included, but the second day is not
        """
        filter_from = None
        filter_to = None

        if stamp_filter is None:
            return filter_from, filter_to

        if '--' in stamp_filter:
            filter_from, filter_to = stamp_filter.split('--')
            filter_from = datetime.strptime(filter_from, self.date_format)
            filter_to = datetime.strptime(filter_to, self.date_format)

        elif stamp_filter.startswith('*'):
            filter_to = datetime.strptime(stamp_filter, '*'+self.date_format)
            filter_to = filter_to.replace(hour=0, minute=0, second=0)

        elif stamp_filter.endswith('*'):
            filter_from = datetime.strptime(stamp_filter, self.date_format+'*')
            filter_from = filter_from.replace(hour=0, minute=0, second=0)

        elif '+' in stamp_filter:
            # "+" filtering works with a date following by "+", a number and
            # a letter (d|w|m|y) which sets the number of days, weeks, months,
            # years we would like to go into the future
            filter_from, number = stamp_filter.split('+')
            filter_from = datetime.strptime(filter_from, self.date_format)
            number, filter_to = self.parse_number_filter(
                number, filter_from, future=True)

        elif stamp_filter.count('-') == 3:
            # "-" filtering works with a date followed by "-", a number and
            # a letter (d|w|m|y) which sets the number of days, weeks, months,
            # years we would like to go backwards into the past
            year, month, day, number = stamp_filter.split('-')
            filter_to = '-'.join([year, month, day])
            filter_to = datetime.strptime(filter_to, self.date_format)
            number, filter_from = self.parse_number_filter(number, filter_to)

        else:
            # Check if the user is asking for N days/weeks/months/years
            number, filter_from = self.parse_number_filter(stamp_filter)
            if number is None:
                # no filtering found, maybe they are giving us a fixed date
                try:
                    filter_from = datetime.strptime(stamp_filter,
                                                    self.date_format)
                except:
                    # nothing to be used as a filter, go on, printing a warning
                    print('[warning] invalid date filter: ' + stamp_filter)
                else:
                    filter_from = filter_from.replace(hour=0, minute=0,
                                                      second=0)
                    filter_to = filter_from + timedelta(days=1)

        return filter_from, filter_to
